"""Game class module.

Contains classes used in the game.
"""

import pygame
from pygame import draw
import random as R
import euclid
import math

import data
from locals import *
from sprite_classes import *

class ImagelessGroup(pygame.sprite.AbstractGroup):
    """Group that contains sprites without images.
       Instead, each sprite added has it's own draw method which is called.
       sprites).""" 
       
    def draw(self, surface):
        """draw(surface)
           draw all sprites onto the surface

           Calls the draw method for each sprite."""       
        spritedict = self.spritedict
        dirty = self.lostsprites
        self.lostsprites = []
        dirty_append = dirty.append
        for s in self.sprites():
            r = spritedict[s]
            #From pygame.sprite.RenderUpdates with 1 changed line
            #If no image, let sprite be responsible for own draw method
            if not s.image:
                newrect = s.draw(surface)
            else:
                newrect = surface.blit(s.image, s.rect)
                          
            if r is 0:
                dirty_append(newrect)
            else:
                if newrect.colliderect(r):
                    dirty_append(newrect.union(r))
                else:
                    dirty_append(newrect)
                    dirty_append(r)
            spritedict[s] = newrect
        return dirty
    

class Block(pygame.sprite.Sprite):
    """The base class for cellular objects.
    
    Defines a common set of operations to use on each object"""
    def __init__(self,
            position=euclid.Point2(0,0),
            size=(0,0),
            velocity=euclid.Vector2(0,0),
            zoomlevel=1,
            color=pygame.color.Color("White")):
        """Block constructor"""
        pygame.sprite.Sprite.__init__(self)
        self.image = None
        
        self.alive = True
        self.zoomlevel = zoomlevel        
        self.color = color
        
        self.position = position#euclid.Point2(position[0], position[1])
        self.velocity = velocity#euclid.Vector2(velocity[0], velocity[1])
        
        self.rect = pygame.Rect(0,0,0,0)
        self.rect.center = (self.position.x,self.position.y)
        self.rect.width = size[0]/self.zoomlevel
        self.rect.height = size[1]/self.zoomlevel

    def is_alive(self):
        """Used to test if an object is still alive."""
        return self.alive
        
    def kill(self):
        pygame.sprite.Sprite.kill(self)
        self.alive = False

    def get_zoomlevel(self):
        return self.zoomlevel

    def get_position(self):
        return self.position

    def get_color(self):
        return self.color
        
    def get_velocity(self):
        return self.velocity

    def get_area(self):
        return self.rect.width * self.rect.height

    def get_size(self):
        """Return the size of an object."""
        return (self.rect.width + self.rect.height)/4

    def set_color(self, color):
        self.color = color
        
    def set_zoomlevel(self, zoomlevel):
        """Adjust the zoomlevel of an object
        
        The zoom level determines the size of an object on the screen
        relative to it's size."""
        print "Possible exponential zoom bug in block.set_zoomlevel"
        self.zoomlevel = zoomlevel
        self.rect.width = self.rect.width/self.zoomlevel
        self.rect.height = self.rect.width/self.zoomlevel
        self.rect.center = self.position

    def set_position(self, position):
        """Set the position of a block"""        
        self.position = position
        self.rect.center = (position.x,position.y)

    def set_velocity(self, velocity):
        self.velocity = velocity

    def consume(self, block): 
        """Consume another block, increasing the current block's size.
        
        The difference in the size of the blocks is equal to the size of the
        increase divided by a constant the represents energy needed for the
        consumption.  The consummed block is killed after it is consummed."""
        diff = ((self.get_size() - (self.get_size() - block.get_size())) / 
            CONSUME_ENERGY)
        self.rect.width += diff / self.zoomlevel
        self.rect.height += diff / self.zoomlevel
        self.rect.center = self.position
        block.kill()
        return diff

    def move_rel(self, pos):
        """Relative movement based on pos."""
        self.position += pos

    def move(self):
        """Simple position based on velocity"""
        self.position += self.velocity
        self.rect.center = (self.position.x,self.position.y)
        
    def draw(self, surface):
        return draw.rect(surface, self.color, self.rect, 1)


class Cell(Block):
    """A block object that includes a radius"""

    touched = False
    hit = False
    def __init__(self,
            position=euclid.Point2(0,0),
            radius=MINR,
            velocity=euclid.Vector2(0,0),
            zoomlevel=1,
            color=pygame.color.Color("White")):
        """Relies on block's init, with the exception of radius.
    
        Radius is the absolute size of a cell, use for scoring, etc."""
        Block.__init__(self,position,(radius*2,radius*2),velocity,zoomlevel,color)
        self.radius = radius 
        
        #print self.copy()       

    def __copy__(self):
        return Cell(self.position.copy(), self.radius, 
                self.velocity.copy(), self.zoomlevel, self.color)
    copy = __copy__
    
    def set_zoomlevel(self, zoomlevel):
        #Block.set_zoomlevel(self,zoomlevel)
        self.zoomlevel = zoomlevel    
        self.rect.width = int(self.radius*2/self.zoomlevel)
        self.rect.height = int(self.radius*2/self.zoomlevel)
        self.rect.center = self.position
        if self.radius/self.zoomlevel < MINR:
            self.kill()           

    def is_hit(self):
        return self.hit
    
    def is_touched(self):
        return self.touched
        
    def set_hit(self, hit):
        self.hit = hit
        
    def set_touched(self, touched):
        self.touched = touched

    def set_radius(self, radius):
        self.radius = radius
        self.rect.width = self.radius*2/self.zoomlevel
        self.rect.height = self.radius*2/self.zoomlevel
        self.rect.center = self.position

    def get_area(self):
        """Return the size of an object."""
        return math.pi * self.radius * self.radius
        
    def get_radius(self):
        return self.get_size()
        
    def get_size(self):
        return self.radius

    def consume(self, cell):
        diff = ((self.get_radius() - (self.get_radius() - cell.get_radius())) / 
            CONSUME_ENERGY)
        self.radius += diff    
        self.rect.width += diff / self.zoomlevel
        self.rect.height += diff / self.zoomlevel
        self.rect.center = self.position
        cell.kill()
        return diff     

    def draw(self, surface):
        size = int(self.radius/self.zoomlevel)
        return draw.circle(surface, self.color, self.rect.center, size, 0)  
        
    def update(self):
        """Simple movement with screen restrictions.
        
        If the cell is outside the screen with a velocity away from the screen
        destroy the object. """        
        Block.move(self)
        
        if (self.position.x+self.radius/self.zoomlevel < 0 and 
                self.velocity.x < 0):
            self.kill()
        if (self.position.x-self.radius/self.zoomlevel > SCREEN_WIDTH and 
                self.velocity.x > 0):
            self.kill()
        if (self.position.y+self.radius/self.zoomlevel < 0 and 
                self.velocity.y < 0):
            self.kill()
        if (self.position.y-self.radius/self.zoomlevel > SCREEN_HEIGHT and 
                self.velocity.y > 0):
            self.kill()     

class Amoeba(Cell):
    """The Hero(sic) of our Game!
    
    Amoeba is a genus of protozoa that moves by means of temporary projections
    called pseudopods, and is well-known as a representative unicellular
    organism... It obtains its food by phagocytosis... Amoebas reproduce through 
    binary fission.
    
    Reference: http://en.wikipedia.org/wiki/Amoeba """
    def __init__(self, radius = START_RADIUS, zoomlevel = 1):
        """Relies on block's init, with the exception of radius.
    
        Radius is the absolute size of a cell, use for scoring, etc.""" 
        Cell.__init__(self, radius= radius*zoomlevel) 
        self.set_position(euclid.Point2(SCREEN_WIDTH/2, SCREEN_HEIGHT/2))
        self.color = pygame.Color("Light Yellow")

    def move(self, pos, rect):
        """Moves to a given position based upon the distance to that position.
        
        The further away a given position is, the faster the velocity""" 
        x,y = pos   
        if pos != self.position:
            self.velocity.x += (x - self.position.x) / SCREEN_WIDTH
            self.velocity.y += (y - self.position.y) / SCREEN_HEIGHT

            self.velocity *= ELASTISTICITY
            Cell.move(self) 
            self.rect.clamp_ip(rect)           

    def set_hit(self, hit): pass        
    def set_touched(self, touched): pass

    def draw(self, surface):
        size = int(self.radius/self.zoomlevel)
        c = draw.circle(surface, self.color, self.rect.center, size, 0) 
        draw.circle(surface, pygame.Color("White"), self.rect.center, size/2, 0)  
        return c

    def kill(self):
        print "Final Player Radius:",self.radius
        print "Press Enter to respawn player in middle of screen."
        Cell.kill(self)

def coin_flip():
    return R.randint(0,1) == 0
    
def random_color():
    return (R.randint(MINC, MAXC), R.randint(MINC, MAXC), R.randint(MINC, MAXC))
        
class Simplecell(Cell): 
    """A single-cell lifeform that travels with a constant velocity.
    
    This class is the simplest enemy object in the game."""  
    def __init__(self, zoomlevel=1):
        """Create a cell with some random variables.
        
        The simplecell is created right outside the border of the screen, with
        a random velocity somewhere the screen, a random randius, and color."""
        radius = R.randint(MINR, MAXR) * zoomlevel
        color = random_color()
        
        p = euclid.Point2(0,0)                      
        if coin_flip():
            if coin_flip():
                p.x = -radius/zoomlevel
            else:
                p.x = SCREEN_WIDTH + radius/zoomlevel
            p.y = R.randint(0,SCREEN_HEIGHT)
        else:
            if coin_flip():
                p.y = -radius/zoomlevel
            else:
                p.y = SCREEN_HEIGHT + radius/zoomlevel
            p.x = R.randint(0, SCREEN_WIDTH) 
        
        v = euclid.Vector2(
            float(R.randint(0,SCREEN_WIDTH) - p.x) / SCREEN_WIDTH,
            float(R.randint(0,SCREEN_HEIGHT) - p.y) / SCREEN_HEIGHT)
        
        #Call the Cell constructor with our setup arguments
        Cell.__init__(self, p, radius ,v,zoomlevel,color)
        self.adjust_size()

    def set_zoomlevel(self, zoomlevel):
        Cell.set_zoomlevel(self, zoomlevel)
        self.adjust_size()

    def consume(self, cell): 
        Cell.consume(self, cell) 
        self.adjust_size()

    def adjust_size(self):
        """Update the size of the rect, for more accurate object collision.
        
        As circle doesn't take up the entire space of it's bounded square."""
        self.rect.width *= .75
        self.rect.height *= .75                     

class Intelligentcell(Cell):
    """A Cell with some level of AI, including avoidance and attraction
       algorithms."""
    def __init__(self,
            position=euclid.Point2(0,0),
            radius=MINR,
            velocity=euclid.Vector2(0,0),
            zoomlevel=1,
            color=pygame.color.Color("White")):       
        Cell.__init__(position,radius,velocity,zoomlevel,color)
        pass


class Splitter(Simplecell): 
    """An object that causes mitosis in the player (split in two)."""
    def __init__(self, zoomlevel=1):
        Simplecell.__init__(self, zoomlevel)
        self.set_radius(6*zoomlevel)

    def draw(self, surface):
        self.color = random_color()
        diamond = ((self.rect.center[0],self.rect.top),
                   (self.rect.right,self.rect.center[1]),
                   (self.rect.center[0],self.rect.bottom),
                   (self.rect.left,self.rect.center[1]))
        return draw.polygon(surface, self.color, diamond, 0) 

class Pseudopod(pygame.sprite.Sprite):
    """A pseudopod like tentacle class
    
    A pseudopod goes from a start object until it reaches an end object.
    Once it reaches it's target, it starts to pull it back to it's start.
    """
    def __init__(self, start, end, max):    
        pygame.sprite.Sprite.__init__(self)
        self.image = None
        self.time = 0 
        self.start = start
        
        self.end = end
        self.max = max
        self.linesize = (self.start.get_radius()/self.start.get_zoomlevel()) /2
        self.color = self.start.get_color()

    def set_zoomlevel(self,zoomlevel):
        pass

    def update(self, *args):
        self.time += 1
        d = self.start.position.distance(self.end.position)
        if d > self.max:
            self.kill()
        if d == 0:
            self.kill()
        if not self.end.alive or not self.start.alive:
            self.kill()

        #The greater the distance, the slower the pull
        #The longer the pull, the faster it increases        
        self.end.velocity.x =+ self.time*self.time/d * (self.start.position.x - self.end.position.x)/self.max
        self.end.velocity.y =+ self.time*self.time/d * (self.start.position.y - self.end.position.y)/self.max        

    def draw(self, surface):
        return draw.line(surface, self.color, 
            self.start.position, self.end.position, int(self.linesize))   
           
