"""Game function module.

Contains functions used in the game.
"""

import pygame
from pygame.locals import *
import euclid
import math
import random as R

from classes import *
from locals import *

def add_enemy(level,zoomlevel):
    
    images = None

    if level==0:
        return Simplecell(zoomlevel)    
    elif level==1:
        images = ['ebola.png','goli.png','virus.png','bacterium.png','cell.png']    
    elif level==2:
        images = ['bug1.png','bug2.png','bug3.png','bug4.png','bug5.png','bug6.png',
                  'bug7.png']
    elif level==3:
        images = ['roach.png','mouse.png','monkey.png','cat.png', 'dog.png'] 
    elif level==4:
        images = ['spacerock.png','satelight.png','ufo.png','ship.png','asteroid.png', 'ss.png']
    elif level==5:
        images = ['venus.png','pluto.png','uranus.png','saturn.png','jupiter.png', 'sun.png'] 
    else:
        return Simplecell(zoomlevel)  
             
    rand = R.randint(0,len(images)-1)
    return SpriteBlock(images[rand], zoomlevel=zoomlevel)


def find_closest(cell,group,minsize,maxsize,maxdistance):
    """Find the closest object in a group under maxdistance which is between
       minsize and maxdize.  If no object exists, none is returned"""
    cr = cell.get_size()

    md = maxdistance
    myg = None
    for g in group:
        gr = g.get_size()
        if minsize < gr and maxsize > gr:
            pos = cell.get_position() 
            d = pos.distance(g.get_position())
            if d < md:
                myg = g
                md = d                
    return myg

def create_vesticle(player, enemies, zoomlevel):
    """Create a vesticle, given several parameters.
    
    A Vesticle is a small projectile cell used to change the vector of another
    cell.  This is done in order to save the cell pinching off the vesticle from
    being consumed.  The vesticle will automatically target the closest cell
    that is larger than it's previous host."""
    vest_size = player.get_size() * VESTICLE_PERCENTAGE

    #Only pinch a vesticle if it won't shrink us to much
    if (player.get_size()/zoomlevel - vest_size/zoomlevel) < START_RADIUS:
        return None

    #Find nearest enemy with a greater radius than myself                    
    enemy = find_closest(player, enemies, player.get_size(),
                               player.get_area(), SCREEN_WIDTH + SCREEN_HEIGHT)
    if not enemy:
        return None

    ppos = player.get_position()
    epos = enemy.get_position()
    #Need a more uniforn vector formula, because vesticle is small, it needs
    #to be traveling quite fast
    x = -(ppos.x - epos.x) / 8
    y = -(ppos.y - epos.y) / 8

    player.set_radius(player.get_size() - vest_size/2)
    return Cell(player.get_position().copy(),vest_size*2,euclid.Vector2(x,y),
                player.get_zoomlevel(), player.get_color())


def collide(a,b,Fa,Fb):
    """Adjust two objects new vectors and positions."""
    a.set_velocity(Fa)
    b.set_velocity(Fb)        
    a.move_rel(Fa)
    b.move_rel(Fa)
    a.set_hit(True)
    b.set_hit(True)

def compare_objects(a,b,ar,br,Fa,Fb):
    """Compare the outcome of two objects based upon their size."""
    if ar > br:
        if ar < br * FUZZ_SIZE:
            collide(a,b,Fa,Fb)
        elif ar > br * TOO_SMALL:
            if not b.is_hit():
                b.set_velocity(Fb)
                b.move_rel(Fb)
                b.set_hit(True)
                #collide(a,b,Fa,Fb)
        else:
            a.consume(b)     

def determine_outcome(a,b):
    """Find the outcome of two items that have collided.
    
    The result is either a vector collision, or one object consumes the other"""    
    Fa, Fb = vector_collision(a,b)
    ar = a.get_size()
    br = b.get_size()
    
    if ar == br:
        collide(a,b,Fa,Fb)
    else:
        compare_objects(a,b,ar,br,Fa,Fb)
        compare_objects(b,a,br,ar,Fb,Fa)

def handle_zoom(zoomlevel, zk, player, all, b):
    newzoom = zoomlevel + (zoomlevel * .05) * zk
    dirty = None

    if player.get_size()/newzoom > MAX_RADIUS and (newzoom < float(zoomlevel)):
        return zoomlevel, dirty

    if ((newzoom >= 1) and (player.get_size()/newzoom >= START_RADIUS)):            
        zoomlevel = newzoom 
        dirty = b.background_zoom(zk)
        for a in all:
            a.set_zoomlevel(zoomlevel)
    return zoomlevel, dirty
     
def vector_collision(objecta,objectb):
    """Computes the final vectors of two objects that collide based upon weight.

    Performs some trickly vector math to compute the final vectors of two
    objects that collide givin the objects initial vectors and their weight.
    Several constants are also used which are defined below.

    See: http://www.euclideanspace.com/physics/dynamics/collision/twod/index.htm   
    """
    #Constants: Everything must be a float for this function to work correctly!    
    ra = euclid.Vector2(0., 0.)
    rb = euclid.Vector2(0., 0.)
    Ia = 1. #Inertia scalar for object a
    Ib = 1. #Inertia scalar or object b
    e = 1.  #e = coefficient of restitution
    
    #taking out the math.pi on the mass dynamically changes the equation, why?
    Vai,ma = objecta.get_velocity(), float(objecta.get_area())#float(objecta.get_size())
    Vbi,mb = objectb.get_velocity(), float(objecta.get_area())#float(objectb.get_size())

    k= 1/(ma*ma)+ 2/(ma*mb) +1/(mb*mb) - \
        (ra.x*ra.x)/(ma*Ia) - (rb.x*rb.x)/(ma*Ib) - (ra.y*ra.y)/(ma*Ia) - \
        (ra.y*ra.y)/(mb*Ia) - (ra.x*ra.x)/(mb*Ia) - (rb.x*rb.x)/(mb*Ib) - \
        (rb.y*rb.y)/(ma*Ib) - (rb.y*rb.y)/(mb*Ib) + \
        (ra.y*ra.y*rb.x*rb.x)/(Ia*Ib) + \
        (ra.x*ra.x*rb.y*rb.y)/(Ia*Ib) - \
        (2*ra.x*ra.y*rb.x*rb.y)/(Ia*Ib)
    
    Jx =  (e+1)/k * (Vai.x - Vbi.x) * \
        ( 1/ma - ra.x*ra.x/Ia + 1/mb - rb.x*rb.x/Ib) - \
        (e+1)/k * (Vai.y - Vbi.y) * (ra.x*ra.y / Ia + rb.x*rb.y / Ib)     
    Jy = -(e+1)/k * (Vai.x - Vbi.x) * \
        (ra.x*ra.y / Ia + rb.x*rb.y / Ib) + (e+1)/k  * (Vai.y - Vbi.y) * \
        ( 1/ma - ra.y*ra.y/Ia + 1/mb - rb.y*rb.y/Ib)    

    #The code on the referenced page appears wrong at the following step
    #Vbf.x = Vbi.x - Jx/Mb; Vbf.y = Vbi.y - Jy/Mb;
    Fa = euclid.Vector2(Vai.x - Jx/ma, Vai.y - Jy/ma)
    Fb = euclid.Vector2(Vbi.x + Jx/mb, Vbi.y + Jy/mb)
 
    return Fa,Fb
    
            
def toggle_fullscreen():
    """Shift between fullscreen and previous screen-size
    
    See http://www.pygame.org/wiki/toggle_fullscreen?parent=CookBook
    """
    screen = pygame.display.get_surface()
    tmp = screen.convert()
    caption = pygame.display.get_caption()
    
    w,h = screen.get_width(),screen.get_height()
    flags = screen.get_flags()
    bits = screen.get_bitsize()
    
    pygame.display.quit()
    pygame.display.init()
    
    screen = pygame.display.set_mode((w,h),flags^FULLSCREEN,bits)
    screen.blit(tmp,(0,0))
    pygame.display.set_caption(*caption)
 
    pygame.key.set_mods(0) #HACK: work-a-round for a SDL bug??
    
    return screen
